use std::mem::size_of;

use zerocopy::{AsBytes, FromBytes, Unaligned};

use crate::{Db, FilePointer, FileRange, PagePointer, PAGE_SIZE, U16, U32};

#[derive(Clone, Copy, PartialEq, Eq, Debug)]
enum SlabKind {
    SingleBytes,
    RelativeFreeList,
    AbsoluteFreeList,
}

impl SlabKind {
    fn for_size(size: u32) -> Self {
        if size == 1 {
            Self::SingleBytes
        } else if size < size_of::<FilePointer>() as u32 {
            Self::RelativeFreeList
        } else if (size as u64) <= PAGE_SIZE {
            Self::AbsoluteFreeList
        } else {
            panic!("invalid size")
        }
    }
}

// Best bitmap sizes in bytes for a 4096 Byte slab
// const BITMAP_SIZE: [u32; 8] = [456, 241, 164, 125, 100, 84, 72, 63];

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct FreeList {
    head: FilePointer,
}

impl FreeList {
    pub fn empty() -> Self {
        Self {
            head: FilePointer::null(),
        }
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct AllocatorState {
    pub general: FreeList,
    pub slabs: SlabListPointer,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct SlabListHeader {
    next: SlabListPointer,
    len: U32,
    size: U32,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct SlabListPointer(pub FilePointer);

pub struct SlabListIterator<'db> {
    position: u32,
    db: &'db Db,
    ptr: SlabListPointer,
}

impl<'db> Iterator for SlabListIterator<'db> {
    type Item = SlabPointer;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(res) = self.ptr.get(self.db, self.position) {
            self.position += 1;
            Some(res)
        } else if let Some(next) = self.ptr.next(self.db) {
            self.ptr = next;
            self.next()
        } else {
            None
        }
    }
}

impl SlabListHeader {
    pub fn capacity(&self) -> u32 {
        (self.size.get() - size_of::<SlabListHeader>() as u32) / size_of::<Slab>() as u32
    }
}

impl SlabListPointer {
    pub fn next(self, db: &Db) -> Option<SlabListPointer> {
        let ptr: SlabListPointer = self.read_header(db).next;

        (!ptr.0.is_null()).then_some(ptr)
    }

    fn read_header(self, db: &Db) -> SlabListHeader {
        unsafe { db.read(self.0) }
    }

    fn modify_header(self, db: &mut Db) -> &mut SlabListHeader {
        unsafe { db.modify(self.0) }
    }

    pub fn set_next(self, db: &mut Db, next: SlabListPointer) {
        self.modify_header(db).next = next;
    }

    pub fn set_len(self, db: &mut Db, len: u32) {
        self.modify_header(db).len = U32::from(len);
    }

    pub fn init(self, db: &mut Db, size: u32) {
        *self.modify_header(db) = SlabListHeader {
            next: SlabListPointer(FilePointer::null()),
            size: size.into(),
            len: 0.into(),
        };
    }

    pub fn ptr(self, db: &Db, i: u32) -> Option<FilePointer> {
        let this = self.read_header(db);
        (i < this.len.get()).then(|| {
            self.0 + size_of::<SlabListHeader>() as u64 + i as u64 * size_of::<Slab>() as u64
        })
    }

    pub fn write(self, db: &mut Db, i: u32, value: Slab) {
        let ptr = self.ptr(db, i).unwrap();
        unsafe { db.write(ptr, value) };
    }

    pub fn get(self, db: &Db, i: u32) -> Option<SlabPointer> {
        self.ptr(db, i).map(SlabPointer)
    }

    pub fn iter(self, db: &Db) -> SlabListIterator {
        SlabListIterator {
            position: 0,
            db,
            ptr: self,
        }
    }

    pub fn add(self, db: &mut Db, slab_size: u32) -> SlabPointer {
        let this = self.read_header(db);
        let capacity = this.capacity();
        let SlabListHeader { mut next, len, .. } = this;

        if len.get() >= capacity {
            if next.0.is_null() {
                next = SlabListPointer(db.add_pages(1).start());
                next.init(db, PAGE_SIZE as u32);
                self.set_next(db, next);
            }

            return next.add(db, slab_size);
        }

        let len = len.get();
        self.set_len(db, len + 1);
        self.write(
            db,
            len,
            Slab {
                head: FilePointer::null(),
                size: slab_size.into(),
            },
        );

        SlabPointer(self.ptr(db, len).unwrap())
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct Slab {
    head: FilePointer,
    size: U32,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
struct RelativeFreeListHeader {
    next_page: PagePointer,
    first: U16,
}

impl RelativeFreeListHeader {
    const DATA_SIZE: u32 = PAGE_SIZE as u32 - size_of::<Self>() as u32;

    fn capacity(size: u32) -> u32 {
        debug_assert_eq!(SlabKind::for_size(size), SlabKind::RelativeFreeList);
        Self::DATA_SIZE / size
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct SlabPointer(FilePointer);

impl SlabPointer {
    fn read(&self, db: &Db) -> Slab {
        unsafe { db.read(self.0) }
    }

    fn modify<'db>(&self, db: &'db mut Db) -> &'db mut Slab {
        unsafe { db.modify(self.0) }
    }

    pub fn alloc(&self, db: &mut Db) -> FileRange {
        let Slab { mut head, size } = self.read(db);

        if head.is_null() {
            head = self.allocate_page(db);
        }

        let size = size.get();

        match SlabKind::for_size(size) {
            SlabKind::SingleBytes => todo!(),
            SlabKind::RelativeFreeList => {
                let (page, offset) = head.page_offset();
                assert_eq!(offset, 0);

                let RelativeFreeListHeader { first, .. } = unsafe { db.read(page.start()) };

                // the page should never be full if its in the free list
                assert_ne!(first.get(), 0);

                let ptr = FilePointer::from_page_offset(page, first.get());

                let next: U16 = unsafe { db.read(ptr) };

                let header = unsafe { db.modify::<RelativeFreeListHeader>(page.start()) };

                header.first = next;

                if next.get() == 0 {
                    // page is full
                    let next_page = header.next_page;
                    header.next_page = PagePointer::null();
                    self.modify(db).head = next_page.start();
                }

                ptr
            }
            SlabKind::AbsoluteFreeList => {
                let next = unsafe { db.read(head) };
                self.set_head(db, next);
                head
            }
        }
        .range(size as u64)
    }

    pub fn free(&self, db: &mut Db, range: FileRange) {
        let Slab { head, size } = self.read(db);

        assert_eq!(range.len(), size.get() as u64);

        let size = size.get();

        match SlabKind::for_size(size) {
            SlabKind::SingleBytes => todo!(),
            SlabKind::RelativeFreeList => {
                let (page, offset) = range.start.page_offset();

                let RelativeFreeListHeader { first, .. } = unsafe { db.read(page.start()) };

                // update next pointer of new element in free list
                unsafe { db.write(range.start, first) };

                let header = unsafe { db.modify::<RelativeFreeListHeader>(page.start()) };

                // point to new element
                header.first = offset.into();

                if first.get() == 0 {
                    // page was full

                    let (head_page, offset) = head.page_offset();
                    assert_eq!(offset, 0);

                    header.next_page = head_page;

                    self.modify(db).head = page.start();
                }
            }
            SlabKind::AbsoluteFreeList => {
                unsafe { db.write(range.start, head) };
                self.set_head(db, range.start);
            }
        }
    }

    pub fn set_head(&self, db: &mut Db, next: FilePointer) {
        self.modify(db).head = next;
    }

    pub fn allocate_page(&self, db: &mut Db) -> FilePointer {
        let Slab { head, size } = self.read(db);

        println!("allocate_page({size})");

        let size = size.get();

        match SlabKind::for_size(size) {
            SlabKind::SingleBytes => todo!(),
            SlabKind::RelativeFreeList => {
                let page = db.add_pages(1);

                let (next_page, offset) = head.page_offset();
                assert_eq!(offset, 0);

                let capacity = RelativeFreeListHeader::capacity(size);

                let data_offset = size_of::<Self>() as u16;

                unsafe {
                    db.write(
                        page.start(),
                        RelativeFreeListHeader {
                            next_page,
                            first: data_offset.into(),
                        },
                    )
                };

                let mut offset = 0;
                for i in (0..capacity).rev() {
                    let next = data_offset + (i * size) as u16;

                    unsafe {
                        db.write(FilePointer::from_page_offset(page, next), U16::from(offset))
                    };

                    offset = next;
                }

                self.set_head(db, page.start());

                page.start()
            }
            SlabKind::AbsoluteFreeList => {
                let n = PAGE_SIZE / size as u64;

                let page = db.add_pages(1);

                let mut next = head;
                for i in (0..n).rev() {
                    let current = page.start() + i * size as u64;
                    unsafe { db.write(current, next) };
                    next = current;
                }

                self.set_head(db, next);

                next
            }
        }
    }
}
